from typing import List

import numpy as np
import numpy.typing as npt

from signaturizer3d.input.sdf import gather_sdf_data
from signaturizer3d.model import FineTunedUniMol
from signaturizer3d.space import CCSpace


class Signaturizer:
    def __init__(self, space: CCSpace):
        if not isinstance(space, CCSpace):
            try:
                space = CCSpace[space]
            except KeyError:
                raise ValueError(
                    f"{space} is not a valid CCSpace. Valid spaces are A1 ... E5, see the CCSpace enum."
                )
        self.space = space

        # Hydrogens are always removed as the signaturizers models are fine
        # tuned from the pre-trained UniMol model trained without hydrogens
        self.model = FineTunedUniMol(space=self.space, remove_hs=True)

    def infer_from_coordinates(
        self, atoms: List[str], coordinates: List[List[float]]
    ) -> npt.NDArray[np.float32]:  # (n, 128) dim array for n signatures
        sig_output = self.model.get_sig4_coordinates(atoms, coordinates)
        return sig_output

    def infer_from_smiles(self, smiles_list: List[str]) -> npt.NDArray[np.float32]:
        sig_output = self.model.get_sig4_smiles(smiles_list)
        return sig_output

    def infer_from_sdf(self, path: str) -> npt.NDArray[np.float32]:
        """
        Perform inference on molecules in an SDF file or all SDF files within a directory.

        Parameters:
        path (str): Path to an SDF file or a directory containing SDF files.
        """
        atoms_list, coords_list = gather_sdf_data(path)
        return self.infer_from_coordinates(atoms_list, coords_list)



# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional

import torch
from torch import Tensor, nn

from .softmax_dropout import softmax_dropout


class SelfMultiheadAttention(nn.Module):
    def __init__(
        self,
        embed_dim,
        num_heads,
        dropout=0.1,
        bias=True,
        scaling_factor=1,
    ):
        super().__init__()
        self.embed_dim = embed_dim

        self.num_heads = num_heads
        self.dropout = dropout

        self.head_dim = embed_dim // num_heads
        assert (
            self.head_dim * num_heads == self.embed_dim
        ), "embed_dim must be divisible by num_heads"
        self.scaling = (self.head_dim * scaling_factor) ** -0.5

        self.in_proj = nn.Linear(embed_dim, embed_dim * 3, bias=bias)
        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias)

    def forward(
        self,
        query,
        key_padding_mask: Optional[Tensor] = None,
        attn_bias: Optional[Tensor] = None,
        return_attn: bool = False,
    ) -> Tensor:
        bsz, tgt_len, embed_dim = query.size()
        assert embed_dim == self.embed_dim

        q, k, v = self.in_proj(query).chunk(3, dim=-1)

        q = (
            q.view(bsz, tgt_len, self.num_heads, self.head_dim)
            .transpose(1, 2)
            .contiguous()
            .view(bsz * self.num_heads, -1, self.head_dim)
            * self.scaling
        )
        if k is not None:
            k = (
                k.view(bsz, -1, self.num_heads, self.head_dim)
                .transpose(1, 2)
                .contiguous()
                .view(bsz * self.num_heads, -1, self.head_dim)
            )
        if v is not None:
            v = (
                v.view(bsz, -1, self.num_heads, self.head_dim)
                .transpose(1, 2)
                .contiguous()
                .view(bsz * self.num_heads, -1, self.head_dim)
            )

        assert k is not None
        src_len = k.size(1)

        # This is part of a workaround to get around fork/join parallelism
        # not supporting Optional types.
        if key_padding_mask is not None and key_padding_mask.dim() == 0:
            key_padding_mask = None

        if key_padding_mask is not None:
            assert key_padding_mask.size(0) == bsz
            assert key_padding_mask.size(1) == src_len

        attn_weights = torch.bmm(q, k.transpose(1, 2))

        assert list(attn_weights.size()) == [bsz * self.num_heads, tgt_len, src_len]

        if key_padding_mask is not None:
            # don't attend to padding symbols
            attn_weights = attn_weights.view(bsz, self.num_heads, tgt_len, src_len)
            attn_weights.masked_fill_(
                key_padding_mask.unsqueeze(1).unsqueeze(2).to(torch.bool), float("-inf")
            )
            attn_weights = attn_weights.view(bsz * self.num_heads, tgt_len, src_len)

        if not return_attn:
            attn = softmax_dropout(
                attn_weights,
                self.dropout,
                self.training,
                bias=attn_bias,
            )
        else:
            attn_weights += attn_bias
            attn = softmax_dropout(
                attn_weights,
                self.dropout,
                self.training,
                inplace=False,
            )

        o = torch.bmm(attn, v)
        assert list(o.size()) == [bsz * self.num_heads, tgt_len, self.head_dim]

        o = (
            o.view(bsz, self.num_heads, tgt_len, self.head_dim)
            .transpose(1, 2)
            .contiguous()
            .view(bsz, tgt_len, embed_dim)
        )
        o = self.out_proj(o)
        if not return_attn:
            return o
        else:
            return o, attn_weights, attn



# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function

import argparse
import logging
import pathlib

import numpy as np
import torch
import torch.nn as nn

from signaturizer3d.unimol.pad import pad_1d_tokens, pad_2d, pad_coords
from signaturizer3d.unicore.dictionary import Dictionary
from signaturizer3d.unicore.transformer_encoder import init_bert_params
from signaturizer3d.unicore.unicore_model import BaseUnicoreModel
from signaturizer3d.unicore.utils import get_activation_fn

from .transformers import TransformerEncoderWithPair

logger = logging.getLogger(__name__)

BACKBONE = {
    "transformer": TransformerEncoderWithPair,
}

WEIGHT_DIR = pathlib.Path(__file__).resolve().parents[2] / "weights"


class UniMolModel(BaseUnicoreModel):
    def __init__(
        self,
        classification_head_name: str,
        remove_hs: bool,
        output_dim: int,
        model_file_name: str = None,
        model_file_URL: str = None,
    ):
        super().__init__()
        self.args = finetuned_architecture()

        self.model_file_name = model_file_name
        self.classification_head_name = classification_head_name
        self.remove_hs = remove_hs
        self.output_dim = output_dim

        current_dir = pathlib.Path(__file__).resolve().parents[0]
        self.dictionary = Dictionary.load((current_dir / "dict.txt").as_posix())

        self.mask_idx = self.dictionary.add_symbol("[MASK]", is_special=True)
        self.padding_idx = self.dictionary.pad()
        self.embed_tokens = nn.Embedding(
            len(self.dictionary), self.args.encoder_embed_dim, self.padding_idx
        )
        self.encoder = BACKBONE[self.args.backbone](
            encoder_layers=self.args.encoder_layers,
            embed_dim=self.args.encoder_embed_dim,
            ffn_embed_dim=self.args.encoder_ffn_embed_dim,
            attention_heads=self.args.encoder_attention_heads,
            emb_dropout=self.args.emb_dropout,
            dropout=self.args.dropout,
            attention_dropout=self.args.attention_dropout,
            activation_dropout=self.args.activation_dropout,
            max_seq_len=self.args.max_seq_len,
            activation_fn=self.args.activation_fn,
            no_final_head_layer_norm=self.args.delta_pair_repr_norm_loss < 0,
        )
        K = 128
        n_edge_type = len(self.dictionary) * len(self.dictionary)
        self.gbf_proj = NonLinearHead(
            K, self.args.encoder_attention_heads, self.args.activation_fn
        )
        if self.args.kernel == "gaussian":
            self.gbf = GaussianLayer(K, n_edge_type)
        else:
            raise ValueError("Current not support kernel: {}".format(self.args.kernel))

        self.classification_heads = nn.ModuleDict()
        self.classification_heads[self.classification_head_name] = ClassificationHead(
            input_dim=self.args.encoder_embed_dim,
            inner_dim=self.args.encoder_embed_dim,
            num_classes=self.output_dim,
            activation_fn=self.args.pooler_activation_fn,
            pooler_dropout=self.args.pooler_dropout,
        )

        self.apply(init_bert_params)
        if model_file_name:
            self.pretrain_path = (WEIGHT_DIR / self.model_file_name).as_posix()
            self.load_pretrained_weights(path=self.pretrain_path)
        elif model_file_URL:
            self.download_and_load_pretrained_weights(model_file_URL)
        else:
            raise ValueError("Please provide either model_file_name or model_file_URL")

    def load_pretrained_weights(self, path):
        logger.info(f"Loading pretrained weights from {path}")
        state_dict = torch.load(path, map_location=lambda storage, loc: storage)
        self.load_state_dict(state_dict["model"], strict=False)

    def download_and_load_pretrained_weights(self, url):
        logger.info(f"Loading pretrained weights from {url}")
        state_dict = torch.hub.load_state_dict_from_url(url)
        self.load_state_dict(state_dict["model"], strict=False)

    @classmethod
    def build_model(cls, args):
        """Build a new model instance."""
        return cls(args)

    def forward(
        self,
        src_tokens,
        src_distance,
        src_coord,
        src_edge_type,
        **kwargs,
    ):
        padding_mask = src_tokens.eq(self.padding_idx)
        if not padding_mask.any():
            padding_mask = None
        x = self.embed_tokens(src_tokens)

        def get_dist_features(dist, et):
            n_node = dist.size(-1)
            gbf_feature = self.gbf(dist, et)
            gbf_result = self.gbf_proj(gbf_feature)
            graph_attn_bias = gbf_result
            graph_attn_bias = graph_attn_bias.permute(0, 3, 1, 2).contiguous()
            graph_attn_bias = graph_attn_bias.view(-1, n_node, n_node)
            return graph_attn_bias

        graph_attn_bias = get_dist_features(src_distance, src_edge_type)
        (
            encoder_rep,
            _,
            _,
            _,
            _,
        ) = self.encoder(x, padding_mask=padding_mask, attn_mask=graph_attn_bias)
        cls_repr = encoder_rep[:, 0, :]  # CLS token repr
        all_repr = encoder_rep[:, :, :]  # all token repr

        filtered_tensors = []
        for tokens in src_tokens:
            filtered_tensor = tokens[
                (tokens != 0) & (tokens != 1) & (tokens != 2)
            ]  # filter out BOS(0), EOS(1), PAD(2)
            filtered_tensors.append(filtered_tensor)

        lengths = [
            len(filtered_tensor) for filtered_tensor in filtered_tensors
        ]  # Compute the lengths of the filtered tensors

        cls_atomic_reprs = []
        for i in range(len(all_repr)):
            atomic_repr = encoder_rep[i, 1 : lengths[i] + 1, :]
            cls_atomic_reprs.append(atomic_repr)

        logits = self.classification_heads[self.classification_head_name](cls_repr)

        return logits

    def batch_collate_fn(self, samples):
        batch = {}
        for k in samples[0][0].keys():
            if k == "src_coord":
                v = pad_coords(
                    [torch.tensor(s[0][k]).float() for s in samples], pad_idx=0.0
                )
            elif k == "src_edge_type":
                v = pad_2d(
                    [torch.tensor(s[0][k]).long() for s in samples],
                    pad_idx=self.padding_idx,
                )
            elif k == "src_distance":
                v = pad_2d(
                    [torch.tensor(s[0][k]).float() for s in samples], pad_idx=0.0
                )
            elif k == "src_tokens":
                v = pad_1d_tokens(
                    [torch.tensor(s[0][k]).long() for s in samples],
                    pad_idx=self.padding_idx,
                )
            batch[k] = v
        try:
            label_np = np.array([s[1] for s in samples])
            label = torch.tensor(label_np)
        except:
            label = None
        return batch, label


class ClassificationHead(nn.Module):
    """Head for sentence-level classification tasks."""

    def __init__(
        self,
        input_dim,
        inner_dim,
        num_classes,
        activation_fn,
        pooler_dropout,
    ):
        super().__init__()
        self.dense = nn.Linear(input_dim, inner_dim)
        self.activation_fn = get_activation_fn(activation_fn)
        self.dropout = nn.Dropout(p=pooler_dropout)
        self.out_proj = nn.Linear(inner_dim, num_classes)

    def forward(self, features, **kwargs):
        x = features
        x = self.dropout(x)
        x = self.dense(x)
        x = self.activation_fn(x)
        x = self.dropout(x)
        x = self.out_proj(x)
        return x


class NonLinearHead(nn.Module):
    """Head for simple classification tasks."""

    def __init__(
        self,
        input_dim,
        out_dim,
        activation_fn,
        hidden=None,
    ):
        super().__init__()
        hidden = input_dim if not hidden else hidden
        self.linear1 = nn.Linear(input_dim, hidden)
        self.linear2 = nn.Linear(hidden, out_dim)
        self.activation_fn = get_activation_fn(activation_fn)

    def forward(self, x):
        x = self.linear1(x)
        x = self.activation_fn(x)
        x = self.linear2(x)
        return x


@torch.jit.script
def gaussian(x, mean, std):
    pi = 3.14159
    a = (2 * pi) ** 0.5
    return torch.exp(-0.5 * (((x - mean) / std) ** 2)) / (a * std)


class GaussianLayer(nn.Module):
    def __init__(self, K=128, edge_types=1024):
        super().__init__()
        self.K = K
        self.means = nn.Embedding(1, K)
        self.stds = nn.Embedding(1, K)
        self.mul = nn.Embedding(edge_types, 1)
        self.bias = nn.Embedding(edge_types, 1)
        nn.init.uniform_(self.means.weight, 0, 3)
        nn.init.uniform_(self.stds.weight, 0, 3)
        nn.init.constant_(self.bias.weight, 0)
        nn.init.constant_(self.mul.weight, 1)

    def forward(self, x, edge_type):
        mul = self.mul(edge_type).type_as(x)
        bias = self.bias(edge_type).type_as(x)
        x = mul * x.unsqueeze(-1) + bias
        x = x.expand(-1, -1, -1, self.K)
        mean = self.means.weight.float().view(-1)
        std = self.stds.weight.float().view(-1).abs() + 1e-5
        return gaussian(x.float(), mean, std).type_as(self.means.weight)


def finetuned_architecture():
    args = argparse.ArgumentParser()
    args.encoder_layers = getattr(args, "encoder_layers", 15)
    args.encoder_embed_dim = getattr(args, "encoder_embed_dim", 512)
    args.encoder_ffn_embed_dim = getattr(args, "encoder_ffn_embed_dim", 2048)
    args.encoder_attention_heads = getattr(args, "encoder_attention_heads", 64)
    args.dropout = getattr(args, "dropout", 0.1)
    args.emb_dropout = getattr(args, "emb_dropout", 0.1)
    args.attention_dropout = getattr(args, "attention_dropout", 0.1)
    args.activation_dropout = getattr(args, "activation_dropout", 0.0)
    args.pooler_dropout = getattr(args, "pooler_dropout", 0)
    args.max_seq_len = getattr(args, "max_seq_len", 512)
    args.activation_fn = getattr(args, "activation_fn", "gelu")
    args.pooler_activation_fn = getattr(args, "pooler_activation_fn", "tanh")
    args.post_ln = getattr(args, "post_ln", False)
    args.backbone = getattr(args, "backbone", "transformer")
    args.kernel = getattr(args, "kernel", "gaussian")
    args.delta_pair_repr_norm_loss = getattr(args, "delta_pair_repr_norm_loss", -1.0)
    return args



import pathlib
import pickle
from typing import TYPE_CHECKING

import torch

if TYPE_CHECKING:
    from signaturizer3d.space import CCSpace

DIR = pathlib.Path(__file__).resolve().parents[0]


def normalize_for_space(raw_outputs: torch.Tensor, space: "CCSpace") -> torch.Tensor:
    means = pickle.load(
        open(
            DIR / "space_means.pkl",
            "rb",
        )
    )
    stds = pickle.load(
        open(
            DIR / "space_stds.pkl",
            "rb",
        )
    )

    space_mean = torch.tensor(means[space.name], device=raw_outputs.device)
    space_std = torch.tensor(stds[space.name], device=raw_outputs.device)
    normalized_output = raw_outputs * space_std + space_mean
    return normalized_output



# Copyright (c) DP Technology.
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import contextlib
import importlib
import logging
import os
import sys
from functools import partial
from typing import Callable, List

import torch
import torch.nn.functional as F
import torch.utils.checkpoint

logger = logging.getLogger(__name__)

try:
    import unicore_fused_multi_tensor

    HAS_MULTI_TENSOR = True
except:
    logger.info("fused_multi_tensor is not installed corrected")
    HAS_MULTI_TENSOR = False

try:
    import unicore_fused_rounding

    HAS_FUSED_ROUNDING = True
except:
    logger.info("fused_rounding is not installed corrected")
    HAS_FUSED_ROUNDING = False

if not torch.cuda.is_available() or torch.cuda.get_device_capability()[0] < 7:
    HAS_MULTI_TENSOR = False
    HAS_FUSED_ROUNDING = False

logger = logging.getLogger(__name__)


def apply_to_sample(f, sample):
    if hasattr(sample, "__len__") and len(sample) == 0:
        return {}

    def _apply(x):
        if torch.is_tensor(x):
            return f(x)
        elif isinstance(x, dict):
            return {key: _apply(value) for key, value in x.items()}
        elif isinstance(x, list):
            return [_apply(x) for x in x]
        elif isinstance(x, tuple):
            return tuple(_apply(x) for x in x)
        elif isinstance(x, set):
            return {_apply(x) for x in x}
        else:
            return x

    return _apply(sample)


def move_to_cuda(sample, device=None):
    device = device or torch.cuda.current_device()

    def _move_to_cuda(tensor):
        # non_blocking is ignored if tensor is not pinned, so we can always set
        # to True (see github.com/PyTorchLightning/pytorch-lightning/issues/620)
        return tensor.to(device=device, non_blocking=True)

    return apply_to_sample(_move_to_cuda, sample)


def move_to_cpu(sample):
    def _move_to_cpu(tensor):
        # PyTorch has poor support for half tensors (float16) on CPU.
        # Move any such tensors to float32.
        if tensor.dtype in {torch.bfloat16, torch.float16}:
            tensor = tensor.to(dtype=torch.float32)
        return tensor.cpu()

    return apply_to_sample(_move_to_cpu, sample)


def multi_tensor_total_norm(grads, chunk_size=2048 * 32) -> torch.Tensor:
    per_device_grads = {}
    norms = []
    for grad in grads:
        device = grad.device
        dtype = grad.dtype
        if device not in per_device_grads:
            per_device_grads[device] = {}
        if dtype not in per_device_grads[device]:
            per_device_grads[device][dtype] = []
        per_device_grads[device][dtype].append(grad)
    for device in per_device_grads.keys():
        for dtype in per_device_grads[device].keys():
            cur_grads = per_device_grads[device][dtype]
            if HAS_MULTI_TENSOR and device.type == "cuda":
                norm = unicore_fused_multi_tensor.l2norm(chunk_size, [cur_grads])
                norms.append(norm)
            else:
                norms += [torch.norm(g, p=2, dtype=torch.float32) for g in cur_grads]
    total_norm = torch.norm(torch.stack(norms), p=2, dtype=torch.float32)
    return total_norm


@torch.no_grad()
def clip_grad_norm_(params, max_norm, aggregate_norm_fn=None) -> torch.Tensor:
    if isinstance(params, torch.Tensor):
        params = [params]
    params = list(params)
    grads = [p.grad.detach() for p in filter(lambda p: p.grad is not None, params)]
    if len(grads) == 0:
        if len(params) > 0:
            return params[0].new_tensor(0.0)
        else:
            return torch.tensor(0.0)

    if len(grads) == 1:
        total_norm = torch.norm(grads[0], p=2, dtype=torch.float32)
    else:
        total_norm = multi_tensor_total_norm(grads)

    if aggregate_norm_fn is not None:
        total_norm = aggregate_norm_fn(total_norm)

    if max_norm > 0:
        max_norm = float(max_norm)
        clip_coef = (max_norm / (total_norm + 1e-6)).clamp_(max=1)
        for g in grads:
            g.mul_(clip_coef)
    return total_norm


def import_user_module(args):
    module_path = getattr(args, "user_dir", None)
    if module_path is not None:
        module_path = os.path.abspath(args.user_dir)
        if not os.path.exists(module_path) and not os.path.isfile(
            os.path.dirname(module_path)
        ):
            unicore_rel_path = os.path.join(os.path.dirname(__file__), args.user_dir)
            if os.path.exists(unicore_rel_path):
                module_path = unicore_rel_path
            else:
                unicore_rel_path = os.path.join(
                    os.path.dirname(__file__), "..", args.user_dir
                )
                if os.path.exists(unicore_rel_path):
                    module_path = unicore_rel_path
                else:
                    raise FileNotFoundError(module_path)

        # ensure that user modules are only imported once
        import_user_module.memo = getattr(import_user_module, "memo", set())
        if module_path not in import_user_module.memo:
            import_user_module.memo.add(module_path)

            module_parent, module_name = os.path.split(module_path)
            if module_name not in sys.modules:
                sys.path.insert(0, module_parent)
                importlib.import_module(module_name)
            else:
                raise ImportError(
                    "Failed to import --user-dir={} because the corresponding module name "
                    "({}) is not globally unique. Please rename the directory to "
                    "something unique and try again.".format(module_path, module_name)
                )


def get_activation_fn(activation: str) -> Callable:
    """Returns the activation function corresponding to `activation`"""

    if activation == "relu":
        return F.relu
    elif activation == "gelu":
        return F.gelu
    elif activation == "tanh":
        return torch.tanh
    elif activation == "linear":
        return lambda x: x
    else:
        raise RuntimeError("--activation-fn {} not supported".format(activation))


def get_available_activation_fns() -> List:
    return [
        "relu",
        "gelu",
        "tanh",
        "linear",
    ]


def has_parameters(module):
    try:
        next(module.parameters())
        return True
    except StopIteration:
        return False


def get_rng_state():
    state = {"torch_rng_state": torch.get_rng_state()}
    if torch.cuda.is_available():
        state["cuda_rng_state"] = torch.cuda.get_rng_state()
    return state


def set_rng_state(state):
    torch.set_rng_state(state["torch_rng_state"])
    if torch.cuda.is_available():
        torch.cuda.set_rng_state(state["cuda_rng_state"])


@contextlib.contextmanager
def torch_seed(seed, *addl_seeds):
    """Context manager which seeds the NumPy PRNG with the specified seed and
    restores the state afterward"""
    if seed is None:
        yield
        return

    def check_seed(s):
        assert type(s) == int or type(s) == np.int32 or type(s) == np.int64

    check_seed(seed)
    if len(addl_seeds) > 0:
        for s in addl_seeds:
            check_seed(s)
        seed = int(hash((seed, *addl_seeds)) % 1e8)
    state = get_rng_state()
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
    try:
        yield
    finally:
        set_rng_state(state)


class CudaEnvironment(object):
    def __init__(self):
        cur_device = torch.cuda.current_device()
        prop = torch.cuda.get_device_properties("cuda:{}".format(cur_device))
        self.name = prop.name
        self.major = prop.major
        self.minor = prop.minor
        self.total_memory_in_GB = prop.total_memory / 1024 / 1024 / 1024

    @staticmethod
    def pretty_print_cuda_env_list(cuda_env_list):
        """
        Given a list of CudaEnviorments, pretty print them
        """
        num_workers = len(cuda_env_list)
        center = "CUDA enviroments for all {} workers".format(num_workers)
        banner_len = 40 - len(center) // 2
        first_line = "*" * banner_len + center + "*" * banner_len
        logger.info(first_line)
        for r, env in enumerate(cuda_env_list):
            logger.info(
                "rank {:3d}: ".format(r)
                + "capabilities = {:2d}.{:<2d} ; ".format(env.major, env.minor)
                + "total memory = {:.3f} GB ; ".format(env.total_memory_in_GB)
                + "name = {:40s}".format(env.name)
            )
        logger.info(first_line)


def csv_str_list(x):
    return x.split(",")


def eval_str_list(x, type=float):
    if x is None:
        return None
    if isinstance(x, str):
        x = eval(x)
    try:
        return list(map(type, x))
    except TypeError:
        return [type(x)]


def eval_str_dict(x, type=dict):
    if x is None:
        return None
    if isinstance(x, str):
        x = eval(x)
    return x


def eval_bool(x, default=False):
    if x is None:
        return default
    try:
        return bool(eval(x))
    except TypeError:
        return default


def checkpoint_sequential(
    functions,
    input,
    enabled=True,
):
    def wrap_tuple(a):
        return (a,) if type(a) is not tuple else a

    def exec(func, a):
        return wrap_tuple(func(*a))

    def get_wrap_exec(func):
        def wrap_exec(*a):
            return exec(func, a)

        return wrap_exec

    input = wrap_tuple(input)

    is_grad_enabled = torch.is_grad_enabled()

    if enabled and is_grad_enabled:
        for func in functions:
            input = torch.utils.checkpoint.checkpoint(get_wrap_exec(func), *input)
    else:
        for func in functions:
            input = exec(func, input)
    return input


def permute_final_dims(tensor: torch.Tensor, inds: List[int]):
    zero_index = -1 * len(inds)
    first_inds = list(range(len(tensor.shape[:zero_index])))
    return tensor.permute(first_inds + [zero_index + i for i in inds])


def flatten_final_dims(t: torch.Tensor, num_dims: int):
    return t.reshape(t.shape[:-num_dims] + (-1,))


def masked_mean(mask, value, dim, eps=1e-10):
    mask = mask.expand(*value.shape)
    return torch.sum(mask * value, dim=dim) / (eps + torch.sum(mask, dim=dim))


def dict_multimap(fn, dicts):
    first = dicts[0]
    new_dict = {}
    for k, v in first.items():
        all_v = [d[k] for d in dicts]
        if type(v) is dict:
            new_dict[k] = dict_multimap(fn, all_v)
        else:
            new_dict[k] = fn(all_v)

    return new_dict


def one_hot(x, num_classes, dtype=torch.float32):
    x_one_hot = torch.zeros(*x.shape, num_classes, dtype=dtype, device=x.device)
    x_one_hot.scatter_(-1, x.long().unsqueeze(-1), 1)
    return x_one_hot


def batched_gather(data, inds, dim=0, num_batch_dims=0):
    assert dim < 0 or dim - num_batch_dims >= 0
    ranges = []
    for i, s in enumerate(data.shape[:num_batch_dims]):
        r = torch.arange(s)
        r = r.view(*(*((1,) * i), -1, *((1,) * (len(inds.shape) - i - 1))))
        ranges.append(r)

    remaining_dims = [slice(None) for _ in range(len(data.shape) - num_batch_dims)]
    remaining_dims[dim - num_batch_dims if dim >= 0 else dim] = inds
    ranges.extend(remaining_dims)
    return data[ranges]


def dict_map(fn, dic, leaf_type):
    new_dict = {}
    for k, v in dic.items():
        if type(v) is dict:
            new_dict[k] = dict_map(fn, v, leaf_type)
        else:
            new_dict[k] = tree_map(fn, v, leaf_type)

    return new_dict


def tree_map(fn, tree, leaf_type):
    if isinstance(tree, dict):
        return dict_map(fn, tree, leaf_type)
    elif isinstance(tree, List):
        return [tree_map(fn, x, leaf_type) for x in tree]
    elif isinstance(tree, tuple):
        return tuple([tree_map(fn, x, leaf_type) for x in tree])
    elif isinstance(tree, leaf_type):
        try:
            return fn(tree)
        except:
            raise ValueError(f"cannot apply {fn} on {tree}.")
    else:
        raise ValueError(f"{type(tree)} not supported")


tensor_tree_map = partial(tree_map, leaf_type=torch.Tensor)


def fp32_to_bf16_sr(t, o):
    if HAS_FUSED_ROUNDING and t.device.type == "cuda":
        unicore_fused_rounding.fp32_to_bf16_sr(t, o)
    else:
        r = (
            torch.rand(size=t.size(), device=t.device, dtype=torch.float32) - 0.5
        ) / 256
        m, e = torch.frexp(t)
        t = t + torch.ldexp(r, e)
        o.data.copy_(t.bfloat16())


def set_jit_fusion_options():
    """Set PyTorch JIT layer fusion options."""
    # flags required to enable jit fusion kernels
    # legacy pytorch fuser
    torch._C._jit_set_profiling_mode(False)
    torch._C._jit_set_profiling_executor(False)
    torch._C._jit_override_can_fuse_on_cpu(True)
    torch._C._jit_override_can_fuse_on_gpu(True)


@contextlib.contextmanager
def validate_with_ema(trainer, ema=False):
    if not ema:
        yield
        return
    _wrapped_model = trainer._wrapped_model
    trainer._wrapped_model = trainer.ema.model_ema
    try:
        yield
    finally:
        trainer._wrapped_model = _wrapped_model



# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging

import torch
import torch.nn.functional as F

logger = logging.getLogger(__name__)

try:
    import unicore_fused_softmax_dropout

    HAS_SOFTMAX = True
except:
    logger.info("fused_softmax is not installed corrected")
    HAS_SOFTMAX = False

if not torch.cuda.is_available() or torch.cuda.get_device_capability()[0] < 7:
    HAS_SOFTMAX = False


class SoftmaxDropoutFast(torch.autograd.Function):
    @staticmethod
    def forward(ctx, is_training, inputs, mask, bias, dropout_prob):
        (
            dropout_results,
            dropout_mask,
            softmax_results,
        ) = unicore_fused_softmax_dropout.forward(
            is_training, inputs, mask, bias, dropout_prob, None
        )
        if is_training:
            ctx.dropout_prob = dropout_prob
            ctx.save_for_backward(softmax_results, dropout_mask)
            ctx.has_bias = bias is not None and bias.requires_grad
            if ctx.has_bias:
                ctx.bias_batch_dim = bias.shape[0]
        return dropout_results

    @staticmethod
    def backward(ctx, grad_output):
        softmax_results, dropout_mask = ctx.saved_tensors
        dropout_prob = ctx.dropout_prob
        grad_output = grad_output.contiguous()
        grad_input = unicore_fused_softmax_dropout.backward(
            grad_output, softmax_results, dropout_mask, dropout_prob
        )
        if ctx.has_bias:
            grad_bias = grad_input.view(
                -1, ctx.bias_batch_dim, grad_input.shape[-2], grad_input.shape[-1]
            ).sum(dim=0)
        else:
            grad_bias = None
        return None, grad_input, None, grad_bias, None


def _check_mask(mask, input):
    try:
        assert mask.dtype == input.dtype, "mask and input must have the same dtype"
        assert len(mask.shape) == len(input.shape), "wrong length of mask.shape"
        assert (
            mask.shape[-3] == 1 or mask.shape[-3] == input.shape[-3]
        ), "mask.shape[-3] must be 1 or input.shape[-3]"
        if mask.shape[-3] == 1:
            assert (
                mask.shape[-2] == 1
            ), "when mask.shape[-3] == 1, mask.shape[-2] must be 1"
        else:
            assert (
                mask.shape[-2] == 1 or mask.shape[-2] == input.shape[-2]
            ), "mask.shape[-2] must be 1 or input.shape[-2]"
        return True
    except:
        return False


def _check_bias(bias, input):
    try:
        assert bias.dtype == input.dtype, "bias and input must have the same dtype"
        assert len(bias.shape) == len(input.shape), "wrong length of bias.shape"
        assert (
            bias.shape[-1] == input.shape[-1]
        ), "bias.shape[-1] must be input.shape[-1]"
        assert (
            bias.shape[-2] == input.shape[-2]
        ), "bias.shape[-2] must be input.shape[-2]"
        len_shape = len(input.shape)
        if len_shape > 3:
            # head dim should be the same
            assert (
                bias.shape[-3] == input.shape[-3]
            ), "bias.shape[-3] must be input.shape[-3]"
            offset = 3
        else:
            offset = 2
        prev_non_one = True
        for i in range(len_shape - offset - 1, -1, -1):
            if prev_non_one:
                assert (
                    bias.shape[i] == input.shape[i] or bias.shape[i] == 1
                ), "bias.shape[{}] must be input.shape[{}] or 1".format(i, i)
            else:
                assert bias.shape[i] == 1, "bias.shape[{}] must be 1".format(i)
            prev_non_one = bias.shape[i] != 1
        return True
    except:
        return False


def softmax_dropout(
    input, dropout_prob, is_training=True, mask=None, bias=None, inplace=True
):
    """softmax dropout, and mask, bias are optional.
    Args:
        input (torch.Tensor): input tensor
        dropout_prob (float): dropout probability
        is_training (bool, optional): is in training or not. Defaults to True.
        mask (torch.Tensor, optional): the mask tensor, use as input + mask . Defaults to None.
        bias (torch.Tensor, optional): the bias tensor, use as input + bias . Defaults to None.

    Returns:
        torch.Tensor: the result after softmax
    """
    input = input.contiguous()
    if not inplace:
        # copy a input for non-inplace case
        input = input.clone()
    if input.is_cuda and HAS_SOFTMAX:
        input_size = input.size()
        if mask is not None:
            if _check_mask(mask, input):
                mask = mask.contiguous().view(-1, mask.shape[-2], mask.shape[-1])
            else:
                input += mask
                mask = None
        if bias is not None:
            if _check_bias(bias, input):
                bias = bias.contiguous().view(-1, input_size[-2], input_size[-1])
            else:
                input += bias
                bias = None
        input = input.view(-1, input_size[-2], input_size[-1])
        if dropout_prob <= 0.0 or input_size[-1] <= 1024:
            return SoftmaxDropoutFast.apply(
                is_training, input, mask, bias, dropout_prob
            ).view(*input_size)
        else:
            return F.dropout(
                SoftmaxDropoutFast.apply(is_training, input, mask, bias, 0.0).view(
                    *input_size
                ),
                p=dropout_prob,
                training=is_training,
            )
    else:
        if mask is not None:
            input += mask
        if bias is not None:
            input += bias
        return F.dropout(F.softmax(input, dim=-1), p=dropout_prob, training=is_training)



from typing import List, Union

import numpy as np
import numpy.typing as npt

from signaturizer3d.input.smiles import generate_conformations, validate_smiles
from signaturizer3d.unicore.dictionary import Dictionary


def pairwise_distance_numpy(coords: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
    # Calculate pairwise distance matrix using broadcasting and vector norm
    diff = coords[:, np.newaxis] - coords[np.newaxis, :]
    dist_matrix = np.sqrt(np.sum(diff**2, axis=-1))
    return dist_matrix.astype(np.float32)


def coordinates_to_unimol(
    atoms: List[str],
    coordinates: Union[npt.NDArray[np.float32], List[List[float]]],
    dictionary: Dictionary,
    max_atoms: int = 256,
    remove_hs: bool = True,
) -> dict:
    """
    Converts atom symbols and coordinates to input in unimol's expected representation.

    Parameters:
    atoms (List[str]): List of atom symbols.
    coordinates (numpy array): (n_atoms, 3) numpy array of x, y, z coordinates for each atom.
    dictionary (Dictionary): A Uni-Core dictionay
    max_atoms (int): Maximum number of atoms to consider in the molecule
    remove_hs (bool): Whether to remove hydrogen atoms from the representation.

    Returns:
    dict: A dictionary with keys `src_tokens`, `src_distance`, `src_coord`, `src_edge_type`.
    """
    coordinates = np.array(coordinates).astype(np.float32)
    atoms_array = np.array(atoms)

    if remove_hs:
        non_hydrogen_mask = atoms_array != "H"
        atoms_array = atoms_array[non_hydrogen_mask]
        coordinates_array = coordinates[non_hydrogen_mask]

    # Randomly select atoms if more than max_atoms
    if len(atoms_array) > max_atoms:
        selected_indices = np.random.choice(len(atoms_array), max_atoms, replace=False)
        atoms_array = atoms_array[selected_indices]
        coordinates_array = coordinates_array[selected_indices]

    # Create source tokens with start and end tokens
    src_tokens = np.array(
        [dictionary.bos()]
        + [dictionary.index(atom) for atom in atoms_array]
        + [dictionary.eos()],
    )

    # Normalize coordinates and add padding for start and end tokens
    src_coord = coordinates_array - coordinates_array.mean(axis=0)
    src_coord = np.pad(src_coord, ((1, 1), (0, 0)), "constant", constant_values=0)

    src_distance = pairwise_distance_numpy(src_coord)

    # Calculate edge type, what does this do?
    # src_edge_type = np.outer(src_tokens, src_tokens).astype(int)
    src_edge_type = src_tokens.reshape(-1, 1) * len(dictionary) + src_tokens.reshape(
        1, -1
    )

    return {
        "src_tokens": src_tokens.astype(int),
        "src_distance": src_distance,
        "src_coord": src_coord,
        "src_edge_type": src_edge_type.astype(int),
    }


def coordinates_list_to_unimol(
    atoms_list: List[List[str]],
    coordinates_list: List[List[List[float]]],
    dictionary: Dictionary,
) -> List[dict]:
    assert (
        len(dictionary) == 31
    ), "Dictionary length was 31 in unimol_tools and should be here too"

    unimol_input = []
    for atoms, coordinates in zip(atoms_list, coordinates_list):
        unimol_input.append(
            coordinates_to_unimol(
                atoms,
                coordinates,
                dictionary,
                max_atoms=256,
                remove_hs=True,
            )
        )
    return unimol_input


def smiles_to_unimol(smiles_list: List[str], dictionary: Dictionary) -> List[dict]:
    valid_smiles = validate_smiles(smiles_list)
    atoms_list, coordinates_list = generate_conformations(valid_smiles)
    unimol_input = coordinates_list_to_unimol(atoms_list, coordinates_list, dictionary)  # type: ignore
    return unimol_input



# Copyright (c) DP Technology.
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import logging

import numpy as np

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Dictionary:
    """A mapping from symbols to consecutive integers"""

    def __init__(
        self,
        *,  # begin keyword-only arguments
        bos="[CLS]",
        pad="[PAD]",
        eos="[SEP]",
        unk="[UNK]",
        extra_special_symbols=None,
    ):
        self.bos_word, self.unk_word, self.pad_word, self.eos_word = bos, unk, pad, eos
        self.symbols = []
        self.count = []
        self.indices = {}
        self.specials = set()
        self.specials.add(bos)
        self.specials.add(unk)
        self.specials.add(pad)
        self.specials.add(eos)

    def __eq__(self, other):
        return self.indices == other.indices

    def __getitem__(self, idx):
        if idx < len(self.symbols):
            return self.symbols[idx]
        return self.unk_word

    def __len__(self):
        """Returns the number of symbols in the dictionary"""
        return len(self.symbols)

    def __contains__(self, sym):
        return sym in self.indices

    def vec_index(self, a):
        return np.vectorize(self.index)(a)

    def index(self, sym):
        """Returns the index of the specified symbol"""
        assert isinstance(sym, str)
        if sym in self.indices:
            return self.indices[sym]
        return self.indices[self.unk_word]

    def special_index(self):
        return [self.index(x) for x in self.specials]

    def add_symbol(self, word, n=1, overwrite=False, is_special=False):
        """Adds a word to the dictionary"""
        if is_special:
            self.specials.add(word)
        if word in self.indices and not overwrite:
            idx = self.indices[word]
            self.count[idx] = self.count[idx] + n
            return idx
        else:
            idx = len(self.symbols)
            self.indices[word] = idx
            self.symbols.append(word)
            self.count.append(n)
            return idx

    def bos(self):
        """Helper to get index of beginning-of-sentence symbol"""
        return self.index(self.bos_word)

    def pad(self):
        """Helper to get index of pad symbol"""
        return self.index(self.pad_word)

    def eos(self):
        """Helper to get index of end-of-sentence symbol"""
        return self.index(self.eos_word)

    def unk(self):
        """Helper to get index of unk symbol"""
        return self.index(self.unk_word)

    @classmethod
    def load(cls, f):
        """Loads the dictionary from a text file with the format:

        ```
        <symbol0> <count0>
        <symbol1> <count1>
        ...
        ```
        """
        d = cls()
        d.add_from_file(f)
        return d

    def add_from_file(self, f):
        """
        Loads a pre-existing dictionary from a text file and adds its symbols
        to this instance.
        """
        if isinstance(f, str):
            try:
                with open(f, "r", encoding="utf-8") as fd:
                    self.add_from_file(fd)
            except FileNotFoundError as fnfe:
                raise fnfe
            except UnicodeError:
                raise Exception(
                    "Incorrect encoding detected in {}, please "
                    "rebuild the dataset".format(f)
                )
            return

        lines = f.readlines()

        for line_idx, line in enumerate(lines):
            try:
                splits = line.rstrip().rsplit(" ", 1)
                line = splits[0]
                field = splits[1] if len(splits) > 1 else str(len(lines) - line_idx)
                if field == "#overwrite":
                    overwrite = True
                    line, field = line.rsplit(" ", 1)
                else:
                    overwrite = False
                count = int(field)
                word = line
                if word in self and not overwrite:
                    logger.info(
                        "Duplicate word found when loading Dictionary: '{}', index is {}.".format(
                            word, self.indices[word]
                        )
                    )
                else:
                    self.add_symbol(word, n=count, overwrite=overwrite)
            except ValueError:
                raise ValueError(
                    "Incorrect dictionary format, expected '<token> <cnt> [flags]'"
                )



from typing import List, Union

import numpy as np
import torch
from torch.utils.data import Dataset

from signaturizer3d.inference import run_inference
from signaturizer3d.input import coordinates_list_to_unimol, smiles_to_unimol
from signaturizer3d.space import CCSpace
from signaturizer3d.unimol import UniMolModel


class MolDataset(Dataset):
    def __init__(self, data, label=None):
        self.data = data
        self.label = label if label is not None else np.zeros((len(data), 1))

    def __getitem__(self, idx):
        return self.data[idx], self.label[idx]

    def __len__(self):
        return len(self.data)


class FineTunedUniMol(object):
    def __init__(
        self,
        space: CCSpace,
        remove_hs: bool = True,
        use_gpu: bool = True,
        use_local_weights: bool = False,
    ):
        self.space = space
        self.device = torch.device(
            "cuda:0" if torch.cuda.is_available() and use_gpu else "cpu"
        )

        model_file_name, model_file_url = None, None
        if use_local_weights:
            model_file_name = f"{self.space.name}_split0.pt"
        else:
            model_file_url = (
                "https://github.com/aksell/test-pytorch-modelhub/releases/download/full-CC/"
                + f"{self.space.name}_split0.pt"
            )

        self.model = UniMolModel(
            model_file_name=model_file_name,
            model_file_URL=model_file_url,
            classification_head_name=self.space.name,
            output_dim=128,
            remove_hs=remove_hs,
        ).to(self.device)

        self.model.eval()
        self.params = {"remove_hs": remove_hs}

    def get_sig4_coordinates(
        self, atoms: List[List[str]], coordinates: List[List[List[float]]]
    ):
        unimol_input = coordinates_list_to_unimol(
            atoms, coordinates, self.model.dictionary
        )
        dataset = MolDataset(unimol_input)
        sig4_output = run_inference(
            self.model, space=self.space, dataset=dataset, device=self.device
        )
        return sig4_output

    def get_sig4_smiles(self, smiles_list: Union[List[str], str]):
        if isinstance(smiles_list, str):
            smiles_list = [smiles_list]
        unimol_input = smiles_to_unimol(smiles_list, self.model.dictionary)
        dataset = MolDataset(unimol_input)
        sig4_output = run_inference(
            self.model, space=self.space, dataset=dataset, device=self.device
        )
        return sig4_output



from enum import Enum


class CCSpace(str, Enum):
    """
    Chemical Checker (CC) spaces

    Space names are defined in https://www.nature.com/articles/s41587-020-0502-7
    """

    A1 = "2D fingerprints"
    A2 = "3D fingerprints"
    A3 = "Scaffolds"
    A4 = "Structural keys"
    A5 = "Physiochemistry"
    B1 = "Mechanism of action"
    B2 = "Metabolic genes"
    B3 = "Crystals"
    B4 = "Binding"
    B5 = "HTS bioassasy"
    C1 = "Small molecule roles"
    C2 = "Small molecule pathways"
    C3 = "Signaling pathways"
    C4 = "Biological processes"
    C5 = "Interactome"
    D1 = "Transcription"
    D2 = "Cancer cell lines"
    D3 = "Chemical genetics"
    D4 = "Morphology"
    D5 = "Cell bioassasy"
    E1 = "Theraupetic areas"
    E2 = "Indications"
    E3 = "Side effects"
    E4 = "Diseases & toxicology"
    E5 = "Drug-drug interactions"

    @classmethod
    def print_spaces(cls):
        print("\n".join(str(space) for space in list(cls)))

    def __str__(self):
        return f"{self.name}: {self.value}"



import logging
import os
from typing import List

from rdkit import Chem

logger = logging.getLogger(__name__)


def parse_sdf(sdf_file: str) -> List[tuple[List[str], List[List[float]]]]:
    """
    Parses an SDF file to extract atoms and their coordinates.

    Parameters:
    sdf_file (str): Path to the SDF file.

    Returns:
    List[Tuple[List[str], List[List[float]]]]: A List of tuples, where each tuple contains a list of atom types
    and a list of their corresponding 3D coordinates for each molecule in the SDF file.
    """
    try:
        molecules = Chem.SDMolSupplier(sdf_file)
    except OSError:
        logger.error(f"Unable to read file {sdf_file}")
        return []

    parsed_data = []

    for mol in molecules:
        if mol is None:  # Handling cases where a molecule can't be read
            logger.warning(f"Unable to read molecule from {sdf_file}")
            continue

        atoms = [atom.GetSymbol() for atom in mol.GetAtoms()]
        # Using GetConformer() with the assumption that the molecule has 3D coordinates set.
        coordinates = [
            list(mol.GetConformer().GetAtomPosition(idx))
            for idx in range(mol.GetNumAtoms())
        ]

        parsed_data.append((atoms, coordinates))

    return parsed_data


def gather_sdf_data(path: str) -> tuple[List[List[str]], List[List[List[float]]]]:
    """
    Gather atom and coordinate data from an SDF file or a directory of SDF files.

    Returns:
    Tuple containing two lists:
        - A list of lists of atom types for each molecule.
        - A list of lists of coordinates for each molecule.
    """
    all_atoms = []
    all_coords = []

    if os.path.isfile(path):
        files = [path]
    elif os.path.isdir(path):
        files = [os.path.join(path, f) for f in os.listdir(path) if f.endswith(".sdf")]
    else:
        raise FileNotFoundError(f"No such file or directory: '{path}'")

    for file in files:
        molecules = parse_sdf(file)
        for atoms, coords in molecules:
            all_atoms.append(atoms)
            all_coords.append(coords)

    return all_atoms, all_coords



import logging
from multiprocessing import Pool
from typing import List

import numpy as np
import numpy.typing as npt
from rdkit import Chem
from rdkit.Chem import AllChem
from tqdm import tqdm

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def validate_smiles(smiles_list: List[str]) -> List[str]:
    valid_smiles = []
    invalid_smiles = []

    for smiles in smiles_list:
        if Chem.MolFromSmiles(smiles) is not None:
            valid_smiles.append(smiles)
        else:
            invalid_smiles.append(smiles)

    if invalid_smiles:
        logger.warning(f"Invalid SMILES excluded: {invalid_smiles}")

    return valid_smiles


def inner_smi2coords(smi: str, seed: int):
    """Calculate 3D coordinates for SMILES string"""
    try:
        mol = Chem.MolFromSmiles(smi)
        mol = AllChem.AddHs(mol)
        atoms = [atom.GetSymbol() for atom in mol.GetAtoms()]

        if not atoms:
            raise ValueError(f"No atoms in molecule: {smi}")

        # Embed 3D coordinates
        res = AllChem.EmbedMolecule(mol, randomSeed=seed)
        if res == 0:
            try:
                AllChem.MMFFOptimizeMolecule(mol)
            except:
                pass  # MMFF optimization failed
            coordinates = mol.GetConformer().GetPositions().astype(np.float32)
        else:
            AllChem.Compute2DCoords(mol)
            coordinates = mol.GetConformer().GetPositions().astype(np.float32)

        return atoms, coordinates

    except Exception as e:
        logger.warning(f"Failed to generate conformers for {smi}, reason: {e}")
        atoms = []
        coordinates = np.zeros((0, 3))
        return atoms, coordinates


def smi2coords_wrapper(args):
    smi, seed = args
    return inner_smi2coords(smi, seed)


def generate_conformations(
    smiles_list: List[str], seed: int = -1
) -> tuple[List[List[str]], List[npt.NDArray[np.float32]]]:
    """
    Generate 3D coordinates for a list of SMILES strings

    Returns:
        atoms_list: list of atom types, [["C", "H", "H"], ["C"]]
        coordinates_list: list of numpy arrays containing the cordinates of a molecule with shape (n_atoms,3)
        [[x1, y1, z1], [x2, y2, z2], ...]
    """

    # Initialize multiprocessing pool and generate conformers
    with Pool() as pool:
        logger.info("Start generating conformers...")
        args_list = [(smi, seed) for smi in smiles_list]
        conformers = list(tqdm(pool.imap(smi2coords_wrapper, args_list)))

    # Log success and failure rates
    failed_conformers = [
        conformer for atoms, conformer in conformers if conformer.shape[0] == 0
    ]
    failed_rate = len(failed_conformers) / len(conformers) if len(conformers) else 0
    logger.info(f"Conformer generation success rate: {1 - failed_rate:.2%}")

    atoms_list = [atoms for atoms, _ in conformers]
    coordinates_list = [coords for _, coords in conformers]

    return atoms_list, coordinates_list



# Copyright (c) DP Technology.
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Base classes for various unicore models.
"""

import logging

import torch.nn as nn

logger = logging.getLogger(__name__)


class BaseUnicoreModel(nn.Module):
    """Base class for unicore models."""

    def __init__(self):
        super().__init__()

    @classmethod
    def add_args(cls, parser):
        """Add model-specific arguments to the parser."""
        pass

    @classmethod
    def build_model(cls, args, task):
        """Build a new model instance."""
        raise NotImplementedError("Model must implement the build_model method")

    def extract_features(self, *args, **kwargs):
        """Similar to *forward* but only return features."""
        return self(*args, **kwargs)

    def load_state_dict(
        self,
        state_dict,
        strict=True,
    ):
        """Copies parameters and buffers from *state_dict* into this module and
        its descendants.

        Overrides the method in :class:`nn.Module`.
        """
        return super().load_state_dict(state_dict, strict)

    def set_num_updates(self, num_updates):
        """State from trainer to pass along to model at every update."""

        def _apply(m):
            if hasattr(m, "set_num_updates") and m != self:
                m.set_num_updates(num_updates)

        self.apply(_apply)



# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import numbers

import torch
from torch.nn import functional as F
from torch.nn import init
from torch.nn.parameter import Parameter

logger = logging.getLogger(__name__)
try:
    import unicore_fused_layernorm
    import unicore_fused_layernorm_backward_gamma_beta

    HAS_LAYER_NORM = True
except:
    logger.info("fused_layer_norm is not installed corrected")
    HAS_LAYER_NORM = False

if not torch.cuda.is_available() or torch.cuda.get_device_capability()[0] < 7:
    HAS_LAYER_NORM = False


class FusedLayerNormFastFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, weight, bias, normalized_shape, eps):
        ctx.normalized_shape = normalized_shape
        ctx.eps = eps
        input = input.contiguous()
        weight = weight.contiguous()
        bias = bias.contiguous()
        output, mean, invvar = unicore_fused_layernorm.forward(
            input, ctx.normalized_shape, weight, bias, ctx.eps
        )
        ctx.save_for_backward(input, weight, bias, mean, invvar)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        input_, weight_, bias_, mean, invvar = ctx.saved_tensors
        grad_input = grad_weight = grad_bias = None
        grad_input = unicore_fused_layernorm.backward(
            grad_output.contiguous(),
            mean,
            invvar,
            input_,
            ctx.normalized_shape,
            weight_,
            bias_,
            ctx.eps,
        )
        grad_weight, grad_bias = unicore_fused_layernorm_backward_gamma_beta.backward(
            grad_output.contiguous(),
            mean,
            invvar,
            input_,
            ctx.normalized_shape,
            weight_,
            bias_,
            ctx.eps,
        )
        return grad_input, grad_weight, grad_bias, None, None


FUSED_LAYER_NORM_SUPPORT_DIM = set(
    [
        64,
        128,
        192,
        256,
        320,
        384,
        512,
        640,
        768,
        1024,
        1280,
        1536,
        1792,
        2048,
        2560,
        5120,
    ]
)


class LayerNorm(torch.nn.Module):
    def __init__(self, normalized_shape, eps=1e-5, elementwise_affine=True):
        super(LayerNorm, self).__init__()
        if isinstance(normalized_shape, numbers.Integral):
            normalized_shape = (normalized_shape,)
        self.normalized_shape = torch.Size(normalized_shape)
        self.eps = eps
        assert elementwise_affine
        self.weight = Parameter(torch.Tensor(*normalized_shape))
        self.bias = Parameter(torch.Tensor(*normalized_shape))
        self.reset_parameters()

        def torch_layer_norm(input):
            return F.layer_norm(
                input,
                self.normalized_shape,
                self.weight.type(input.dtype),
                self.bias.type(input.dtype),
                self.eps,
            )

        def fused_layer_norm(input):
            if input.is_cuda:
                return FusedLayerNormFastFunction.apply(
                    input,
                    self.weight.type(input.dtype),
                    self.bias.type(input.dtype),
                    self.normalized_shape,
                    self.eps,
                )
            else:
                return F.layer_norm(
                    input,
                    self.normalized_shape,
                    self.weight.type(input.dtype),
                    self.bias.type(input.dtype),
                    self.eps,
                )

        self.func = (
            torch_layer_norm
            if (
                not HAS_LAYER_NORM
                or normalized_shape[0] not in FUSED_LAYER_NORM_SUPPORT_DIM
            )
            else fused_layer_norm
        )

    def reset_parameters(self):
        init.ones_(self.weight)
        init.zeros_(self.bias)

    def forward(self, input):
        return self.func(input)

    def extra_repr(self):
        return "{normalized_shape}, eps={eps}, " "elementwise_affine=True".format(
            **self.__dict__
        )



# Copyright (c) DP Technology.
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch.nn as nn


def init_bert_params(module):
    if not getattr(module, "can_global_init", True):
        return

    def normal_(data):
        data.copy_(data.cpu().normal_(mean=0.0, std=0.02).to(data.device))

    if isinstance(module, nn.Linear):
        normal_(module.weight.data)
        if module.bias is not None:
            module.bias.data.zero_()
    if isinstance(module, nn.Embedding):
        normal_(module.weight.data)
        if module.padding_idx is not None:
            module.weight.data[module.padding_idx].zero_()



# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from signaturizer3d.unicore.layer_norm import LayerNorm
from signaturizer3d.unicore.transformer_encoder_layer import TransformerEncoderLayer


class TransformerEncoderWithPair(nn.Module):
    def __init__(
        self,
        encoder_layers: int = 6,
        embed_dim: int = 768,
        ffn_embed_dim: int = 3072,
        attention_heads: int = 8,
        emb_dropout: float = 0.1,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        activation_dropout: float = 0.0,
        max_seq_len: int = 256,
        activation_fn: str = "gelu",
        post_ln: bool = False,
        no_final_head_layer_norm: bool = False,
    ) -> None:
        super().__init__()
        self.emb_dropout = emb_dropout
        self.max_seq_len = max_seq_len
        self.embed_dim = embed_dim
        self.attention_heads = attention_heads
        self.emb_layer_norm = LayerNorm(self.embed_dim)
        if not post_ln:
            self.final_layer_norm = LayerNorm(self.embed_dim)
        else:
            self.final_layer_norm = None

        if not no_final_head_layer_norm:
            self.final_head_layer_norm = LayerNorm(attention_heads)
        else:
            self.final_head_layer_norm = None

        self.layers = nn.ModuleList(
            [
                TransformerEncoderLayer(
                    embed_dim=self.embed_dim,
                    ffn_embed_dim=ffn_embed_dim,
                    attention_heads=attention_heads,
                    dropout=dropout,
                    attention_dropout=attention_dropout,
                    activation_dropout=activation_dropout,
                    activation_fn=activation_fn,
                    post_ln=post_ln,
                )
                for _ in range(encoder_layers)
            ]
        )

    def forward(
        self,
        emb: torch.Tensor,
        attn_mask: Optional[torch.Tensor] = None,
        padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        bsz = emb.size(0)
        seq_len = emb.size(1)
        x = self.emb_layer_norm(emb)
        x = F.dropout(x, p=self.emb_dropout, training=self.training)
        # account for padding while computing the representation
        if padding_mask is not None:
            x = x * (1 - padding_mask.unsqueeze(-1).type_as(x))
        input_attn_mask = attn_mask
        input_padding_mask = padding_mask

        def fill_attn_mask(attn_mask, padding_mask, fill_val=float("-inf")):
            if attn_mask is not None and padding_mask is not None:
                # merge key_padding_mask and attn_mask
                attn_mask = attn_mask.view(x.size(0), -1, seq_len, seq_len)
                attn_mask.masked_fill_(
                    padding_mask.unsqueeze(1).unsqueeze(2).to(torch.bool),
                    fill_val,
                )
                attn_mask = attn_mask.view(-1, seq_len, seq_len)
                padding_mask = None
            return attn_mask, padding_mask

        assert attn_mask is not None
        attn_mask, padding_mask = fill_attn_mask(attn_mask, padding_mask)
        for i in range(len(self.layers)):
            x, attn_mask, _ = self.layers[i](
                x, padding_mask=padding_mask, attn_bias=attn_mask, return_attn=True
            )

        def norm_loss(x, eps=1e-10, tolerance=1.0):
            x = x.float()
            max_norm = x.shape[-1] ** 0.5
            norm = torch.sqrt(torch.sum(x**2, dim=-1) + eps)
            error = torch.nn.functional.relu((norm - max_norm).abs() - tolerance)
            return error

        def masked_mean(mask, value, dim=-1, eps=1e-10):
            return (
                torch.sum(mask * value, dim=dim) / (eps + torch.sum(mask, dim=dim))
            ).mean()

        x_norm = norm_loss(x)
        if input_padding_mask is not None:
            token_mask = 1.0 - input_padding_mask.float()
        else:
            token_mask = torch.ones_like(x_norm, device=x_norm.device)
        x_norm = masked_mean(token_mask, x_norm)

        if self.final_layer_norm is not None:
            x = self.final_layer_norm(x)

        delta_pair_repr = attn_mask - input_attn_mask
        delta_pair_repr, _ = fill_attn_mask(delta_pair_repr, input_padding_mask, 0)
        attn_mask = (
            attn_mask.view(bsz, -1, seq_len, seq_len).permute(0, 2, 3, 1).contiguous()
        )
        delta_pair_repr = (
            delta_pair_repr.view(bsz, -1, seq_len, seq_len)
            .permute(0, 2, 3, 1)
            .contiguous()
        )

        pair_mask = token_mask[..., None] * token_mask[..., None, :]
        delta_pair_repr_norm = norm_loss(delta_pair_repr)
        delta_pair_repr_norm = masked_mean(
            pair_mask, delta_pair_repr_norm, dim=(-1, -2)
        )

        if self.final_head_layer_norm is not None:
            delta_pair_repr = self.final_head_layer_norm(delta_pair_repr)

        return x, attn_mask, delta_pair_repr, x_norm, delta_pair_repr_norm



from typing import TYPE_CHECKING

import numpy as np
import torch
from torch.utils.data import DataLoader

from signaturizer3d.normalize.normalize import normalize_for_space

if TYPE_CHECKING:
    from signaturizer3d.model import MolDataset, UniMolModel
    from signaturizer3d.space import CCSpace


def run_inference(
    model: "UniMolModel",
    space: "CCSpace",
    dataset: "MolDataset",
    device,
    batch_size: int = 32,
):
    model = model.to(device)

    dataloader = DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        collate_fn=model.batch_collate_fn,
        shuffle=False,
        drop_last=False,
    )
    model = model.eval()
    raw_outputs_list = []
    for batch in dataloader:
        input_data, _ = batch
        input_data = {k: v.to(device) for k, v in input_data.items()}
        with torch.no_grad():
            raw_outputs = model(**input_data, features_only=True)
            normalized_output = normalize_for_space(raw_outputs, space=space)
            raw_outputs_list.append(normalized_output.cpu().numpy())

    raw_outputs_array = np.concatenate(raw_outputs_list, axis=0)
    return raw_outputs_array



# Copyright (c) DP Technology.
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

from signaturizer3d.unicore import utils
from signaturizer3d.unicore.layer_norm import LayerNorm
from signaturizer3d.unicore.multihead_attention import SelfMultiheadAttention


class TransformerEncoderLayer(nn.Module):
    """
    Implements a Transformer Encoder Layer used in BERT/XLM style pre-trained
    models.
    """

    def __init__(
        self,
        embed_dim: int = 768,
        ffn_embed_dim: int = 3072,
        attention_heads: int = 8,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        activation_dropout: float = 0.0,
        activation_fn: str = "gelu",
        post_ln=False,
    ) -> None:
        super().__init__()

        # Initialize parameters
        self.embed_dim = embed_dim
        self.attention_heads = attention_heads
        self.attention_dropout = attention_dropout

        self.dropout = dropout
        self.activation_dropout = activation_dropout
        self.activation_fn = utils.get_activation_fn(activation_fn)

        self.self_attn = SelfMultiheadAttention(
            self.embed_dim,
            attention_heads,
            dropout=attention_dropout,
        )
        # layer norm associated with the self attention layer
        self.self_attn_layer_norm = LayerNorm(self.embed_dim)
        self.fc1 = nn.Linear(self.embed_dim, ffn_embed_dim)
        self.fc2 = nn.Linear(ffn_embed_dim, self.embed_dim)
        self.final_layer_norm = LayerNorm(self.embed_dim)
        self.post_ln = post_ln

    def forward(
        self,
        x: torch.Tensor,
        attn_bias: Optional[torch.Tensor] = None,
        padding_mask: Optional[torch.Tensor] = None,
        return_attn: bool = False,
    ) -> torch.Tensor:
        """
        LayerNorm is applied either before or after the self-attention/ffn
        modules similar to the original Transformer implementation.
        """
        residual = x
        if not self.post_ln:
            x = self.self_attn_layer_norm(x)
        # new added
        x = self.self_attn(
            query=x,
            key_padding_mask=padding_mask,
            attn_bias=attn_bias,
            return_attn=return_attn,
        )
        if return_attn:
            x, attn_weights, attn_probs = x
        x = F.dropout(x, p=self.dropout, training=self.training)
        x = residual + x
        if self.post_ln:
            x = self.self_attn_layer_norm(x)

        residual = x
        if not self.post_ln:
            x = self.final_layer_norm(x)
        x = self.fc1(x)
        x = self.activation_fn(x)
        x = F.dropout(x, p=self.activation_dropout, training=self.training)
        x = self.fc2(x)
        x = F.dropout(x, p=self.dropout, training=self.training)
        x = residual + x
        if self.post_ln:
            x = self.final_layer_norm(x)
        if not return_attn:
            return x
        else:
            return x, attn_weights, attn_probs



# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


def pad_1d_tokens(
    values,
    pad_idx,
    left_pad=False,
    pad_to_length=None,
    pad_to_multiple=1,
):
    """Convert a list of 1d tensors into a padded 2d tensor."""
    size = max(v.size(0) for v in values)
    size = size if pad_to_length is None else max(size, pad_to_length)
    if pad_to_multiple != 1 and size % pad_to_multiple != 0:
        size = int(((size - 0.1) // pad_to_multiple + 1) * pad_to_multiple)
    res = values[0].new(len(values), size).fill_(pad_idx)

    def copy_tensor(src, dst):
        assert dst.numel() == src.numel()
        dst.copy_(src)

    for i, v in enumerate(values):
        copy_tensor(v, res[i][size - len(v) :] if left_pad else res[i][: len(v)])
    return res


def pad_2d(
    values,
    pad_idx,
    left_pad=False,
    pad_to_length=None,
    pad_to_multiple=1,
):
    """Convert a list of 1d tensors into a padded 2d tensor."""
    size = max(v.size(0) for v in values)
    size = size if pad_to_length is None else max(size, pad_to_length)
    if pad_to_multiple != 1 and size % pad_to_multiple != 0:
        size = int(((size - 0.1) // pad_to_multiple + 1) * pad_to_multiple)
    res = values[0].new(len(values), size, size).fill_(pad_idx)

    def copy_tensor(src, dst):
        assert dst.numel() == src.numel()
        dst.copy_(src)

    for i, v in enumerate(values):
        copy_tensor(
            v,
            res[i][size - len(v) :, size - len(v) :]
            if left_pad
            else res[i][: len(v), : len(v)],
        )
    return res


def pad_coords(
    values,
    pad_idx,
    left_pad=False,
    pad_to_length=None,
    pad_to_multiple=1,
):
    """Convert a list of 1d tensors into a padded 2d tensor."""
    size = max(v.size(0) for v in values)
    size = size if pad_to_length is None else max(size, pad_to_length)
    if pad_to_multiple != 1 and size % pad_to_multiple != 0:
        size = int(((size - 0.1) // pad_to_multiple + 1) * pad_to_multiple)
    res = values[0].new(len(values), size, 3).fill_(pad_idx)

    def copy_tensor(src, dst):
        assert dst.numel() == src.numel()
        dst.copy_(src)

    for i, v in enumerate(values):
        copy_tensor(v, res[i][size - len(v) :, :] if left_pad else res[i][: len(v), :])
    return res
